import configparser
from pathlib import Path

from .make_config_file import make_config_file, DB_CONFIG_FILEPATH


def read_config_file(filepath: Path = DB_CONFIG_FILEPATH) -> dict:
    """
    TODO: docstring
    """

    config = configparser.ConfigParser()
    config.read(filepath)

    all_hosts = {}

    for host in config.sections():
        all_hosts[host] = {}

        for key in config[host]:
            value = config[host][key]

            all_hosts[host][key] = value

    return all_hosts


def configurations(filepath: Path = DB_CONFIG_FILEPATH, verbose: bool = False) -> dict:
    """
    TODO: docstring
    """
    if not filepath.exists():
        make_config_file(filepath)

    if verbose:
        print(f"Loading db configurations from {filepath}")

    return read_config_file(filepath)



from pathlib import Path
from typing import Union

from pathlib import Path

DB_CONFIG_FILEPATH = Path.home() / ".pg-data-etl" / "database_connections.cfg"

STARTER_CONFIG_FILE = """
[DEFAULT]
pw = this-is-a-placeholder-password
port = 5432
super_db = postgres
super_un = postgres
super_pw = this-is-another-placeholder-password

[localhost]
host = localhost
un = postgres
pw = your-password-here

[digitalocean]
un = your-username-here
host = your-host-here.db.ondigitalocean.com
pw = your-password-here
port = 98765
sslmode = require
super_db = your_default_db
super_un = your_super_admin
super_pw = some_super_password12354
"""


def make_config_file(
    filepath: Union[Path, str] = DB_CONFIG_FILEPATH, overwrite: bool = False
) -> bool:
    """
    TODO: docstring
    """

    filepath = Path(filepath)

    # Make sure the parent fiolder exists
    if not filepath.parent.exists():
        filepath.parent.mkdir(parents=True)

    if not overwrite and filepath.exists():
        print("Config file already exists and overwrite=False. Will not overwrite.")
        return False

    else:
        if filepath.exists():
            print(f"Overwriting config file at {filepath}")
        else:
            print(f"Creating a new config file at {filepath}")

        with open(filepath, "w") as open_file:
            open_file.write(STARTER_CONFIG_FILE)

        return True



def table_rename_column(self, old_colname: str, new_colname: str, tablename: str) -> None:
    """
    - Change a column name for a table in SQL
    """

    query = f"ALTER TABLE {tablename} RENAME {old_colname} TO {new_colname};"
    self.execute(query)


def table_add_uid_column(self, tablename: str, uid_col: str = "uid") -> None:
    """
    - Add a unique ID column to a table as a SERIAL PRIMARY KEY type
    """

    query = f"""
        ALTER TABLE {tablename}
        DROP COLUMN IF EXISTS {uid_col};
        
        ALTER TABLE {tablename}
        ADD {uid_col} serial PRIMARY KEY;
    """
    self.execute(query)



from pathlib import Path

from pg_data_etl import helpers


def dump(self, output_folder: Path) -> Path:
    """
    - Create a standalone text file backup of the entire database.
    - Returns the full filepath to the newly created file.
    """

    db_name = self.connection_params["db_name"]
    timestamp = helpers.timestamp_for_filepath()

    filename = f"{db_name}_{timestamp}.sql"
    output_filepath = output_folder / filename

    command = f'{self.cmd.pg_dump} --no-owner --no-acl {self.uri} > "{output_filepath}"'

    print(command)

    helpers.run_command_in_shell(command)

    return output_filepath



import click
from pg_data_etl.settings.make_config_file import make_config_file as _make_config_file


@click.group()
def main():
    "'pg' is used command-line access to the pg_data_etl library"
    pass


@click.command()
@click.option("--overwrite/--no-overwrite", default=False)
def make_config_file(overwrite):
    """Make a configuration file from the template"""
    _make_config_file(overwrite=overwrite)


all_commands = [make_config_file]

for cmd in all_commands:
    main.add_command(cmd)



from .query import *  # noqa
from .admin import *  # noqa
from .backup import *  # noqa
from .copy import *  # noqa
from .data_export import *  # noqa
from .import_geo_data import *  # noqa
from .import_tabular_data import *  # noqa
from .reports import *  # noqa



from pathlib import Path
import pandas as pd
import sqlalchemy
import geopandas as gpd
from geoalchemy2 import Geometry, WKTElement

from pg_data_etl import helpers


def shp2pgsql(self, filepath: str, srid: int, sql_tablename: str, new_srid: int = None):
    """
    Use the shp2pgsql command to import a shapefile into the database
    """

    # Ensure that the schema provided in the 'tablename' exists
    schema, _ = helpers.convert_full_tablename_to_parts(sql_tablename)
    self.add_schema(schema)

    # If 'new_srid' is provided, use 'old:new' to project on the fly
    srid_arg = f"{srid}:{new_srid}" if new_srid else srid

    command = (
        f'{self.cmd.shp2pgsql} -I -s {srid_arg} "{filepath}" {sql_tablename} | psql {self.uri}'
    )

    print(command)

    helpers.run_command_in_shell(command)

    self.gis_table_lint_geom_colname(sql_tablename)


def import_geofile_with_geopandas(
    self, filepath: Path, sql_tablename: str, gpd_kwargs: dict = {}
) -> None:

    # Read the data into a geodataframe
    gdf = gpd.read_file(filepath)

    # Drop null geometries
    gdf = gdf[gdf["geometry"].notnull()]

    self.import_geodataframe(gdf, sql_tablename, gpd_kwargs)


def import_geodataframe(
    self,
    gdf: gpd.GeoDataFrame,
    tablename: str,
    gpd_kwargs: dict = {},
    uid_col: str = "uid",
) -> None:
    """
    TODO: option to use multipart features instead of exploding to singlepart
    """

    gdf = gdf.copy()

    gdf = helpers.sanitize_df_for_sql(gdf)

    epsg_code = int(str(gdf.crs).split(":")[1])

    # Get a list of all geometry types in the dataframe
    geom_types = list(gdf.geometry.geom_type.unique())

    # If there are multi- and single-part features, explode to singlepart
    if len(geom_types) > 1:
        # Explode multipart to singlepart and reset the index
        gdf = gdf.explode()
        gdf["explode"] = gdf.index
        gdf = gdf.reset_index()

    # Use the non-multi version of the geometry
    geom_type_to_use = min(geom_types, key=len).upper()

    # Replace the 'geom' column with 'geometry'
    if "geom" in gdf.columns:
        gdf["geometry"] = gdf["geom"]
        gdf.drop("geom", 1, inplace=True)

    # Drop the 'gid' column
    if "gid" in gdf.columns:
        gdf.drop("gid", 1, inplace=True)

    # Rename 'uid' to 'old_uid'
    if uid_col in gdf.columns:
        gdf[f"old_{uid_col}"] = gdf[uid_col]
        gdf.drop(uid_col, 1, inplace=True)

    # Build a 'geom' column using geoalchemy2
    # and drop the source 'geometry' column
    gdf["geom"] = gdf["geometry"].apply(lambda x: WKTElement(x.wkt, srid=epsg_code))
    gdf.drop("geometry", 1, inplace=True)

    # Ensure that the target schema exists
    schema, tbl = helpers.convert_full_tablename_to_parts(tablename)
    self.schema_add(schema)

    # Write geodataframe to SQL database
    engine = sqlalchemy.create_engine(self.uri)
    gdf.to_sql(
        tbl,
        engine,
        schema=schema,
        dtype={"geom": Geometry(geom_type_to_use, srid=epsg_code)},
        **gpd_kwargs,
    )
    engine.dispose()

    self.table_add_uid_column(tablename)
    self.gis_table_add_spatial_index(tablename)


def import_gis(self, method="geopandas", **kwargs):
    """
    - Import GIS data using `geopandas` or `shp2pgsql`
    - Both methods take `filepath` and `sql_tablename` keyword arguments
    - The geopandas method accepts `gpd_kwargs=dict`
    - The `shp2pgsql` method requires `srid=int` and accepts an optional `new_srid=int` to convert projections during the import process
    """
    method_mapper = {
        "geopandas": import_geofile_with_geopandas,
        "shp2pgsql": shp2pgsql,
    }

    if method not in method_mapper:
        print(f"{method=} does not exist. Valid options include: {method_mapper.keys()}")

    func = method_mapper[method]

    func(self, **kwargs)



"""
`pg_data_etl.database.helpers`
------------------------------

TODO: docstring

"""
from .commands import *  # noqa
from .files import *  # noqa
from .sql_tables import *  # noqa
from .uri import *  # noqa



import psycopg2


def query_as_list_of_lists(self, query: str, super_uri: bool = False) -> list:
    """
    - Use `psycopg2` to run a query and return the result as a list of lists
    - This will NOT commit any changes to the database
    """

    if super_uri:
        uri = self.uri_superuser
    else:
        uri = self.uri

    connection = psycopg2.connect(uri)
    cursor = connection.cursor()

    cursor.execute(query)

    result = cursor.fetchall()

    cursor.close()
    connection.close()

    return [list(x) for x in result]


def query_as_list_of_singletons(self, query: str, super_uri: bool = False):
    """
    - Run a query where the expected output is a list of values
    """

    result = self.query_as_list_of_lists(query, super_uri=super_uri)

    return [x[0] for x in result]


def query_as_singleton(self, query: str, super_uri: bool = False):
    """
    - Run a query where the expected output is a single value
    """

    result = self.query_as_list_of_singletons(query, super_uri=super_uri)

    return result[0]


def exists(self) -> bool:
    """
    - True or False: does this database exist already?
    """

    db_name = self.connection_params["db_name"]

    query = f"""
        SELECT EXISTS(
            SELECT datname FROM pg_catalog.pg_database
            WHERE lower(datname) = lower('{db_name}')
        );
    """

    return self.query_as_singleton(query, super_uri=True)



import psycopg2
import geopandas as gpd


def gdf(self, query: str, geom_col: str = "geom") -> gpd.GeoDataFrame:
    """
    - Get a `geopandas.GeoDataFrame` from a SQL query

    Args:
        query (str): `PostGIS` query as a string
        geom_col (str): geometry column name in the query. Usually `'geom'` or `'shape'`

    Returns:
        query output as GIS data
    """

    connection = psycopg2.connect(self.uri)

    gdf = gpd.GeoDataFrame.from_postgis(query, connection, geom_col=geom_col)

    connection.close()

    return gdf


from typing import Union
from pg_data_etl import helpers


def tables(self, spatial_only: bool = False, schema: Union[str, None] = None) -> list:
    """
    - Return a list of tables in the database
    - Set `spatial_only=True` if you only want a list of geotables
    """
    if spatial_only:
        query = """
            SELECT concat(f_table_schema, '.', f_table_name )
            FROM geometry_columns
        """

        if schema:
            query += f" WHERE f_table_schema = '{schema}'"

    else:
        query = """
            SELECT concat(table_schema, '.', table_name )
            FROM information_schema.tables
            WHERE table_schema NOT IN ('pg_catalog', 'information_schema')
        """

        if schema:
            query += f" AND table_schema = '{schema}'"

    return self.query_as_list_of_singletons(query)


def schemas(self) -> list:
    """
    - Get a list of all schemas in the db
    """

    query = """
        SELECT schema_name
        FROM information_schema.schemata;
    """

    return self.query_as_list_of_singletons(query)


def columns(self, tablename: str) -> list:
    """
    - Get a list of all column names in a given table
    """

    schema, tbl = helpers.convert_full_tablename_to_parts(tablename)

    query = f"""
        SELECT DISTINCT column_name
        FROM information_schema.columns
        WHERE
            table_name = '{tbl}'
            AND
            table_schema = '{schema}';
    """

    return self.query_as_list_of_singletons(query)



from datetime import datetime


def timestamp_for_filepath(dt: datetime = None) -> str:
    """
    Make a datetime string formatted like: 'on_2021_02_09_at_09_29_58'

        TODO: docstring
    """
    if not dt:
        dt = datetime.now()

    return dt.strftime("on_%Y_%m_%d_at_%H_%M_%S")



from pg_data_etl import helpers


def report_spatial(self, print_output: bool = False) -> dict:
    query = "select concat(f_table_schema, '.', f_table_name), srid, type from geometry_columns"

    results = self.query_as_list_of_lists(query)

    output = {}

    for row in results:
        tbl, epsg, geom_type = row

        if epsg not in output:
            output[epsg] = {}

        if geom_type not in output[epsg]:
            output[epsg][geom_type] = []

        output[epsg][geom_type].append(tbl)

    if print_output:

        print("-" * 80)
        print(f"Spatial Data Report for DB: {self.uri}")

        epsg_list = [x for x in output.keys()]
        if len(epsg_list) < 2:
            print(f"\t-> All data is stored in {epsg_list[0]}")
        else:
            print(f"\t-> Data is stored in {len(epsg_list)} projections: {epsg_list}")

        for k in output.keys():
            print(f"\t-> EPSG: {k}")
            for geom_type in output[k]:
                print(f"\t\t->{geom_type}")
                for tbl in output[k][geom_type]:
                    print(f"\t\t\t-> {tbl}")

    return output


def projection(self, tablename: str) -> int:
    """
    - Get the projection of a spatial table

    Args:
        tablename (str): name of table to check, optionally with schema prefix

    Returns:
        EPSG of table
    """
    schema, tbl = helpers.convert_full_tablename_to_parts(tablename)

    query = f"""
        select srid
        from geometry_columns
        where f_table_schema = '{schema}'
        and f_table_name = '{tbl}'
    """

    return self.query_as_singleton(query)



import psycopg2


def execute(self, query: str) -> None:
    """ Use psycopg2 to execute a query & commit it to the database """

    connection = psycopg2.connect(self.uri)
    cursor = connection.cursor()

    cursor.execute(query)

    cursor.close()
    connection.commit()
    connection.close()

    return None



from __future__ import annotations
import subprocess
from pathlib import Path


def run_command_in_shell(command: str) -> str:
    """
    - Use subprocess to execute a command in a shell

    Args:
        command (str)

    TODO: docstring
    """

    process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)

    output = process.communicate()
    print(output[0])

    process.kill()

    return output


class CommandPathManager:
    """
    TODO: docstring
    """

    def __init__(self, psql_bin: str | None = None, tippecanoe_bin: str | None = None):
        self.bins = {"psql": psql_bin, "tippecanoe": tippecanoe_bin}

    def _add_bin_path_if_exists(self, cmd, bin: str = "psql") -> Path:
        """
        TODO: docstring
        """

        if bin in self.bins.keys():
            if not self.bins[bin]:
                return cmd
            else:
                return Path(self.bins[bin]) / cmd
        else:

            print(f"{bin=} is not an option. Use one of: {self.bins.keys()}")
            return None

    def set_bin_path(self, **kwargs) -> None:
        """
        - assign bin path to command using `cmd=bin_path` syntax
        """
        for k, v in kwargs.items():
            self.bins[k] = v

    @property
    def psql(self):
        return self._add_bin_path_if_exists("psql")

    @property
    def pg_dump(self):
        return self._add_bin_path_if_exists("pg_dump")

    @property
    def shp2pgsql(self):
        return self._add_bin_path_if_exists("shp2pgsql")

    @property
    def pgsql2shp(self):
        return self._add_bin_path_if_exists("pgsql2shp")

    @property
    def ogr2ogr(self):
        return self._add_bin_path_if_exists("ogr2ogr")

    @property
    def tippecanoe(self):
        return self._add_bin_path_if_exists("tippecanoe", "tippecanoe")
